"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.getConnectionDirectiveTypeDefs = exports.Connection = exports.Edge = exports.PageInfo = void 0;
const utils_1 = require("@graphql-tools/utils");
exports.PageInfo = `
  type PageInfo {
    hasNextPage: Boolean!
    hasPreviousPage: Boolean!
    startCursor: String
    endCursor: String
  }`;
const Edge = (typeName, args) => {
    const cacheControl = composeCacheContolDirective(args);
    return `
  type ${typeName}Edge {
    cursor: String!
    node: ${typeName.replace('NonNull', '')}${args?.NonNull === true ? '!' : ''} ${cacheControl}
  }
`;
};
exports.Edge = Edge;
const Connection = (typeName, args) => {
    const cacheControl = composeCacheContolDirective(args);
    return `
  type ${typeName}Connection ${cacheControl} {
    totalCount: Int!
    edges: [${typeName}Edge] ${cacheControl}
    pageInfo: PageInfo! ${cacheControl}
  }`;
};
exports.Connection = Connection;
const composeCacheContolDirective = (args) => {
    if (args?.useApolloInheritMaxAge)
        return '@cacheControl(inheritMaxAge: true)';
    if (args?.hasOwnProperty('maxAge'))
        return `@cacheControl(maxAge: ${args.maxAge})`;
    return '';
};
const getConnectionDirectiveTypeDefs = (directiveName) => {
    return `directive @${directiveName || 'connection'} on FIELD_DEFINITION`;
};
exports.getConnectionDirectiveTypeDefs = getConnectionDirectiveTypeDefs;
// Find connection directives
exports.default = (directiveName, cacheControlOptions) => {
    const name = directiveName || 'connection';
    const objectTypes = {};
    const useCacheControl = cacheControlOptions?.enable ?? Boolean(cacheControlOptions);
    const useApolloInheritMaxAge = !!cacheControlOptions?.apollo;
    return {
        connectionObjects: () => objectTypes,
        connectionDirectiveTypeDefs: (0, exports.getConnectionDirectiveTypeDefs)(name),
        connectionDirectiveTransformer: (schema) => {
            const mapper = (fieldConfig) => {
                const directives = (0, utils_1.getDirectives)(schema, fieldConfig);
                if (!directives?.length)
                    return fieldConfig;
                const type = fieldConfig.astNode?.type;
                const args = { NonNull: type.kind === 'NonNullType' };
                const typeName = `${type?.name?.value || type?.type?.name?.value}${args.NonNull ? 'NonNull' : ''}`;
                if (!typeName)
                    return fieldConfig;
                const connectionDirective = directives.find((value) => value.name === name);
                if (!connectionDirective)
                    return fieldConfig;
                if (!useCacheControl) {
                    // Skip cacheContol directives
                    if (!objectTypes.hasOwnProperty(typeName))
                        objectTypes[typeName] = args;
                    return fieldConfig;
                }
                if (useApolloInheritMaxAge) {
                    // Rely on inheritMaxAge instead of calculating maxAge (Apollo v3+)
                    if (!objectTypes.hasOwnProperty(typeName))
                        objectTypes[typeName] = { ...args, useApolloInheritMaxAge };
                    return fieldConfig;
                }
                const cacheControlDirective = directives.find((value) => value.name === 'cacheControl');
                // Fallback to legacy cacheControl calculation (when inheritMaxAge is not supported)
                const cacheControlArgs = cacheControlDirective?.args || {};
                if (!objectTypes.hasOwnProperty(typeName))
                    objectTypes[typeName] = {
                        ...args,
                        ...cacheControlArgs,
                    };
                else if (cacheControlArgs?.maxAge > objectTypes[typeName]?.maxAge)
                    objectTypes[typeName].maxAge = cacheControlArgs.maxAge;
                return fieldConfig;
            };
            return (0, utils_1.mapSchema)(schema, {
                [utils_1.MapperKind.INTERFACE_FIELD]: mapper,
                [utils_1.MapperKind.OBJECT_FIELD]: mapper,
            });
        },
    };
};
